import datetime
import json
from pprint import pprint

from deep_serialization.memory_objects import PythonSerializer
from deep_serialization.json_converter import MemoryDumpEncoder
from deep_serialization.deep_serialization import deserialize_objects


class B:
    def __init__(self, b1, b2, b3):
        self.b1 = b1
        self.b2 = b2
        self.b3 = b3
        self.time = datetime.datetime.now()


def run():
    from pprint import pprint

    c = ["Alex"]
    b = B(1, 2, 3)
    b.b1 = B(4, 5, b)
    a = [1, 2, float('inf'), "abc", {1: 1}, None, b, c]
    serializer_ = PythonSerializer()
    pprint(serializer_.write_object_to_memory(a))
    pprint(serializer_.memory.objects)
    with open('test_json.json', 'w') as fout:
        print(json.dumps(serializer_.memory, cls=MemoryDumpEncoder, indent=True), file=fout)

def deserialize():
    run()
    with open('test_json.json', 'r') as fin:
        data = fin.read()
    pprint(deserialize_objects(["140239390887040"], data, ['copyreg._reconstructor', 'deep_serialization.example.B', 'datetime.datetime']))


if __name__ == '__main__':
    run()



import argparse
import logging

from utbot_executor.listener import PythonExecuteServer


def main(hostname: str, port: int):
    server = PythonExecuteServer(hostname, port)
    server.run()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog = 'UtBot Python Executor',
        description = 'Listen socket stream and execute function value',
        )
    parser.add_argument('hostname')
    parser.add_argument('port', type=int)
    parser.add_argument('--logfile', default=None)
    parser.add_argument(
            '--loglevel',
            choices=["DEBUG", "INFO", "ERROR"],
            default="ERROR",
            )
    args = parser.parse_args()

    loglevel = {"DEBUG": logging.DEBUG, "INFO": logging.INFO, "ERROR": logging.ERROR}[args.loglevel]
    logging.basicConfig(
            filename=args.logfile,
            format='%(asctime)s | %(levelname)s | %(funcName)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=loglevel,
            )
    main(args.hostname, args.port)



import importlib
import json
import logging
import sys
from typing import Dict, Iterable, List, Optional, Set, Union
from utbot_executor.deep_serialization.memory_objects import (
    MemoryObject,
    ReprMemoryObject,
    ListMemoryObject,
    DictMemoryObject,
    ReduceMemoryObject,
    MemoryDump
)
from utbot_executor.deep_serialization.utils import PythonId


class MemoryObjectEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, MemoryObject):
            base_json = {
                'strategy': o.strategy,
                'id': o.id_value(),
                'kind': o.kind,
                'comparable': o.comparable,
            }
            if isinstance(o, ReprMemoryObject):
                base_json['value'] = o.value
            elif isinstance(o, (ListMemoryObject, DictMemoryObject)):
                base_json['items'] = o.items
            elif isinstance(o, ReduceMemoryObject):
                base_json['constructor'] = o.constructor
                base_json['args'] = o.args
                base_json['state'] = o.state
                base_json['listitems'] = o.listitems
                base_json['dictitems'] = o.dictitems
            return base_json
        return json.JSONEncoder.default(self, o)


class MemoryDumpEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, MemoryDump):
            return {id_: MemoryObjectEncoder().default(o) for id_, o in o.objects.items()}
        return json.JSONEncoder.default(self, o)


def as_repr_object(dct: Dict) -> Union[MemoryObject, Dict]:
    if 'strategy' in dct:
        obj: MemoryObject
        if dct['strategy'] == 'repr':
            obj = ReprMemoryObject.__new__(ReprMemoryObject)
            obj.value = dct['value']
            obj.kind = dct['kind']
            obj.comparable = dct['comparable']
            return obj
        if dct['strategy'] == 'list':
            obj = ListMemoryObject.__new__(ListMemoryObject)
            obj.items = dct['items']
            obj.kind = dct['kind']
            obj.comparable = dct['comparable']
            return obj
        if dct['strategy'] == 'dict':
            obj = DictMemoryObject.__new__(DictMemoryObject)
            obj.items = dct['items']
            obj.kind = dct['kind']
            obj.comparable = dct['comparable']
            return obj
        if dct['strategy'] == 'reduce':
            obj = ReduceMemoryObject.__new__(ReduceMemoryObject)
            obj.constructor = dct['constructor']
            obj.args = dct['args']
            obj.state = dct['state']
            obj.listitems = dct['listitems']
            obj.dictitems = dct['dictitems']
            obj.kind = dct['kind']
            obj.comparable = dct['comparable']
            return obj
    return dct


def deserialize_memory_objects(memory_dump: str) -> MemoryDump:
    parsed_data = json.loads(memory_dump, object_hook=as_repr_object)
    return MemoryDump(parsed_data['objects'])


class DumpLoader:
    def __init__(self, memory_dump: MemoryDump):
        self.memory_dump = memory_dump
        self.memory: Dict[PythonId, object] = {}  # key is new id, value is real object

    def add_syspaths(self, syspaths: Iterable[str]):
        for path in syspaths:
            if path not in sys.path:
                sys.path.append(path)

    def add_imports(self, imports: Iterable[str]):
        for module in imports:
            for i in range(1, module.count('.') + 2):
                submodule_name = '.'.join(module.split('.', maxsplit=i)[:i])
                globals()[submodule_name] = importlib.import_module(submodule_name)

    def load_object(self, python_id: PythonId) -> object:
        if python_id in self.memory:
            return self.memory[python_id]

        dump_object = self.memory_dump.objects[python_id]
        real_object: object
        if isinstance(dump_object, ReprMemoryObject):
            real_object = eval(dump_object.value)
        elif isinstance(dump_object, ListMemoryObject):
            if dump_object.kind == 'builtins.set':
                real_object = set(self.load_object(item) for item in dump_object.items)
            elif dump_object.kind == 'builtins.tuple':
                real_object = tuple(self.load_object(item) for item in dump_object.items)
            else:
                real_object = [self.load_object(item) for item in dump_object.items]
        elif isinstance(dump_object, DictMemoryObject):
            real_object = {
                    self.load_object(key): self.load_object(value)
                    for key, value in dump_object.items.items()
                    }
        elif isinstance(dump_object, ReduceMemoryObject):
            constructor = eval(dump_object.constructor)
            args = self.load_object(dump_object.args)
            real_object = constructor(*args)

            self.memory[PythonId(str(id(real_object)))] = real_object
            state = self.load_object(dump_object.state)
            for field, value in state.items():
                setattr(real_object, field, value)
            listitems = self.load_object(dump_object.listitems)
            for listitem in listitems:
                real_object.append(listitem)
            dictitems = self.load_object(dump_object.dictitems)
            for key, dictitem in dictitems.items():
                real_object[key] = dictitem
        else:
            raise TypeError(f'Invalid type {dump_object}')

        self.memory[PythonId(str(id(real_object)))] = real_object
        return real_object


def main():
    with open('test_json.json', 'r') as fin:
        data = fin.read()
    memory_dump = deserialize_memory_objects(data)
    loader = DumpLoader(memory_dump)
    loader.add_imports(['copyreg._reconstructor', 'deep_serialization.example.B', 'datetime.datetime', 'builtins.int', 'builtins.float', 'builtins.bool', 'types.NoneType', 'builtins.list', 'builtins.dict', 'builtins.str', 'builtins.tuple', 'builtins.bytes', 'builtins.type'])
    print(loader.load_object("140239390887040"))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['utbot_executor', 'utbot_executor.deep_serialization']

package_data = \
{'': ['*']}

install_requires = \
['coverage>=6.5.0,<7.0.0']

entry_points = \
{'console_scripts': ['utbot-executor = utbot_executor:utbot_executor']}

setup_kwargs = {
    'name': 'utbot-executor',
    'version': '1.1.16',
    'description': '',
    'long_description': '# UtBot Executor\n\nUtil for python code execution and state serialization.\n\n## Installation\n\nYou can install module from [PyPI](https://pypi.org/project/utbot-executor/):\n\n```bash\npython -m pip install utbot-executor\n```\n\n## Usage\n\n### From console with socket listener\n\nRun with your `<hostname>` and `<port>` for socket connection\n```bash\n$ python -m utbot_executor <hostname> <port> <logfile> [<loglevel DEBUG | INFO | ERROR>]\n```\n\n### Result format:\n\n```json\n{\n        "status": "success",\n        "isException": bool,\n        "statements": list[int],\n        "missedStatements": list[int],\n        "stateBefore": memory json dump,\n        "stateAfter": memory json dump,\n        "argsIds": list[str],\n        "kwargs": list[str],\n        "resultId": str,\n}\n```\n\nor error format:\n\n```json\n{\n        "status": "fail",\n        "exception": str (traceback),\n}\n```\n\n#### States format\n\nTODO\n\n### Submodule `deep_serialization`\n\nJSON serializer and deserializer for python objects\n\n## Source\n\nGitHub [repository](https://github.com/tamarinvs19/utbot_executor)\n',
    'author': 'Vyacheslav Tamarin',
    'author_email': 'vyacheslav.tamarin@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
from typing import Any, Dict, Tuple, List

from utbot_executor.deep_serialization.memory_objects import PythonSerializer
from utbot_executor.deep_serialization.json_converter import MemoryDumpEncoder, deserialize_memory_objects, DumpLoader
from utbot_executor.deep_serialization.utils import PythonId


def serialize_object(obj: Any) -> Tuple[str, str]:
    """
    Serialize one object.
    Returns the object id and memory dump.
    """

    serializer = PythonSerializer()
    id_ = serializer.write_object_to_memory(obj)
    return id_, json.dumps({'objects': serializer.memory}, cls=MemoryDumpEncoder)


def serialize_objects(objs: List[Any]) -> Tuple[List[PythonId], str]:
    """
    Serialize objects with shared memory.
    Returns list of object ids and memory dump.
    """

    serializer = PythonSerializer()
    ids = [
        serializer.write_object_to_memory(obj)
        for obj in objs
    ]
    return ids, json.dumps({'objects': serializer.memory}, cls=MemoryDumpEncoder)


def deserialize_objects(ids: List[str], memory: str, imports: List[str]) -> Dict[str, object]:
    """
    Deserialize objects from shared memory.
    Returns dictionary where keys are ID and values are deserialized objects.
    """

    memory_dump = deserialize_memory_objects(memory)
    loader = DumpLoader(memory_dump)
    return {python_id: loader.load_object(PythonId(python_id)) for python_id in ids}



"""Python code executor for UnitTestBot"""
import inspect
import importlib
import logging
import sys
import traceback
from typing import Any, Callable, Dict, Iterable, List, Set, Tuple
import coverage

from utbot_executor.deep_serialization.deep_serialization import serialize_objects
from utbot_executor.deep_serialization.json_converter import DumpLoader, deserialize_memory_objects
from utbot_executor.deep_serialization.utils import PythonId, getattr_by_path

from utbot_executor.parser import ExecutionRequest, ExecutionResponse, ExecutionFailResponse, ExecutionSuccessResponse
from utbot_executor.utils import suppress_stdout


class PythonExecutor:
    def add_syspaths(self, syspaths: Iterable[str]):
        for path in syspaths:
            if path not in sys.path:
                sys.path.append(path)

    def add_imports(self, imports: Iterable[str]):
        for module in imports:
            for i in range(1, module.count('.') + 2):
                submodule_name = '.'.join(module.split('.', maxsplit=i)[:i])
                globals()[submodule_name] = importlib.import_module(submodule_name)

    def run_function(self, request: ExecutionRequest) -> ExecutionResponse:
        logging.debug("Prepare to run function `%s`", request.function_name)
        logging.debug("Module `%s`", request.function_module)
        logging.debug("Imports `%s`", request.imports)
        try:
            memory_dump = deserialize_memory_objects(request.serialized_memory)
            loader = DumpLoader(memory_dump)
        except Exception as ex:
            logging.debug("Error \n%s", traceback.format_exc())
            return ExecutionFailResponse("fail", traceback.format_exc())

        try:
            self.add_syspaths(request.syspaths)
            self.add_imports(request.imports)
            loader.add_syspaths(request.syspaths)
            loader.add_imports(request.imports)
        except Exception as ex:
            logging.debug("Error \n%s", traceback.format_exc())
            return ExecutionFailResponse("fail", traceback.format_exc())

        try:
            function: Callable = getattr_by_path(
                    importlib.import_module(request.function_module),
                    request.function_name
                    )
            args = [loader.load_object(PythonId(arg_id)) for arg_id in request.arguments_ids]
            kwargs: dict[str, Any] = {}
        except Exception as ex:
            logging.debug("Error \n%s", traceback.format_exc())
            return ExecutionFailResponse("fail", traceback.format_exc())

        try:
            value = run_calculate_function_value(
                    function,
                    args,
                    kwargs,
                    request.filepath
                    )
        except Exception as ex:
            logging.debug("Error \n%s", traceback.format_exc())
            return ExecutionFailResponse("fail", traceback.format_exc())
        logging.debug("Function execution: done")
        return value


def __get_lines(start, end, lines):
    return [x for x in lines if start < x < end]


def _serialize_state(
        args: List[Any],
        kwargs: Dict[str, Any],
        result: Any = None,
        ) -> Tuple[List[PythonId], List[PythonId], PythonId, str]:
    """Serialize objects from args, kwargs and result.

    Returns: tuple of args ids, kwargs ids, result id and serialized memory."""

    all_arguments = args + list(kwargs.values()) + [result]
    ids, serialized_memory = serialize_objects(all_arguments)
    return (
            ids[:len(args)],
            ids[len(args):len(args)+len(kwargs)],
            ids[-1],
            serialized_memory,
            )


def run_calculate_function_value(
        function: Callable,
        args: List[Any],
        kwargs: Dict[str, Any],
        fullpath: str,
    ) -> ExecutionResponse:
    """ Calculate function evaluation result.

    Return serialized data: status, coverage info, object ids and memory."""

    _, _, _, state_before = _serialize_state(args, kwargs)

    __is_exception = False
    __cov = coverage.Coverage(
        data_file=None,
    )
    __cov.start()
    try:
        with suppress_stdout():
            __result = function(*args, **kwargs)
    except Exception as __exception:
        __result = __exception
        __is_exception = True
    __cov.stop()

    (__sources, __start, ) = inspect.getsourcelines(function)
    __end = __start + len(__sources)
    (_, __stmts, _, __missed, _, ) = __cov.analysis2(fullpath)
    __cov.erase()
    __stmts_filtered = __get_lines(__start, __end, __stmts)
    __stmts_filtered_with_def = [__start] + __stmts_filtered
    __missed_filtered = __get_lines(__start, __end, __missed)

    args_ids, kwargs_ids, result_id, state_after = _serialize_state(args, kwargs, __result)

    return ExecutionSuccessResponse(
            status="success",
            is_exception=__is_exception,
            statements=__stmts_filtered_with_def,
            missed_statements=__missed_filtered,
            state_before=state_before,
            state_after=state_after,
            args_ids=args_ids,
            kwargs_ids=kwargs_ids,
            result_id=result_id,
            )
